import re

from mihto.lexer.token_types import TOKEN_PATTERNS


class Lexer:

    def __init__(self, code):
        self.code = code

    def tokenize(self):
        tokens = []
        while self.code:
            tokens.append(self.single_tokenize())
            self.code = self.code.strip()
        return tokens

    def single_tokenize(self):
        for token_type, regex in TOKEN_PATTERNS:
            regex = re.compile(r'\A({})'.format(regex))
            value = regex.search(self.code)
            if value:
                value = value.group(0)
                self.code = self.code[len(value):]
                return Token(token_type, value)

        raise RuntimeError("Couldn't match token on {}".format(self.code))


class Token:
    __slots__ = ('type', 'value')

    def __init__(self, type, value):
        self.type = type
        self.value = value

    def __repr__(self):
        return "<Token type={}, value={} >".format(self.type, self.value)



import mihto.lexer.token_types as ttypes


class ExpressionNode:
    __slots__ = ('value1', 'operator', 'value2')
    _pretty_ops = {ttypes.LESS_EQ_THAN: '<=', ttypes.LESS_THAN: '<',
                   ttypes.GREATER_EQ_THAN: '>=', ttypes.GREATER_THAN: '>',
                   ttypes.EQUALS: '=', ttypes.NOT_EQUALS: '!=', ttypes.AND: 'and', ttypes.OR: 'or'}

    def __init__(self, value1, operator, value2):
        self.value1 = value1
        self.operator = operator
        self.value2 = value2

    def __repr__(self):
        return "<Expression '{} {} {}' >".format(repr(self.value1), repr(self.operator), repr(self.value2))

    def __str__(self):
        return "{} {} {}".format(str(self.value1), self._pretty_ops[self.operator], str(self.value2))


class ValueNode:
    __slots__ = ('value',)
    __nodetype__ = 'valuenode'

    def __init__(self, value):
        self.value = value

    def __repr__(self):
        return "<{} {} >".format(self.__nodetype__, self.value)

    def __str__(self):
        return "{}".format(self.value)


class FloatNode(ValueNode):
    __nodetype__ = "Float"

    def __init__(self, value: float):
        super().__init__(value)


class IntegerNode(ValueNode):
    __nodetype__ = "Integer"

    def __init__(self, value: int):
        super().__init__(value)


class VarRefNode(ValueNode):
    __nodetype__ = "VarRef"

    def __init__(self, value: str):
        super().__init__(value)



from mihto import Lexer, Parser, Evaluator


class Mihto:

    def __init__(self, env=None):
        self.env = env or {}

    def evaluate(self, expression: str):
        lexer = Lexer(expression)
        tokens = lexer.tokenize()
        parser = Parser(tokens)
        expression = parser.parse()
        evaluator = Evaluator(self.env)
        return evaluator.evaluate(expression)



from mihto.parser.nodes import ExpressionNode
import mihto.lexer.token_types as ttypes


class Evaluator:
    OPS = {
        ttypes.AND: lambda x, y: x and y,
        ttypes.OR: lambda x, y: x or y,
        ttypes.LESS_EQ_THAN: lambda x, y: x <= y,
        ttypes.LESS_THAN: lambda x, y: x < y,
        ttypes.GREATER_EQ_THAN: lambda x, y: x >= y,
        ttypes.GREATER_THAN: lambda x, y: x > y,
        ttypes.EQUALS: lambda x, y: x == y,
        ttypes.NOT_EQUALS: lambda x, y: x != y
    }

    def __init__(self, env: dict):
        self.env = env

    def evaluate(self, expression: ExpressionNode) -> bool:
        operator = expression.operator

        if operator in ttypes.LOGICAL_OPERATORS:
            return Evaluator.OPS[operator](self.evaluate(expression.value1), self.evaluate(expression.value2))

        if type(expression.value1) is ExpressionNode or type(expression.value2) is ExpressionNode:
            raise RuntimeError("Operator mismatch. Cannot perform {} operator with boolean data".format(operator))

        value1, value2 = expression.value1.value, expression.value2.value

        if type(value1) is str:
            value1 = self.env.get(value1, None)
            if value1 is None:
                raise RuntimeError("{} cannot be found in environment".format(value1))
        if type(value2) is str:
            value2 = self.env.get(value2, None)
            if value2 is None:
                raise RuntimeError("{} cannot be found in environment".format(value2))

        return Evaluator.OPS[operator](value1, value2)



LESS_EQ_THAN = 'less_eq_than'
GREATER_EQ_THAN = 'greater_eq_than'
LESS_THAN = 'less_than'
GREATER_THAN = 'greater_than'
EQUALS = 'equals'
NOT_EQUALS = 'not_equals'
AND = 'and'
OR = 'or'
IDENTIFIER = 'identifier'
FLOAT = 'float'
INTEGER = 'integer'
OPENPAR = 'openpar'
CLOSEPAR = 'closepar'

TOKEN_PATTERNS = [
    [LESS_EQ_THAN, r"<="],
    [GREATER_EQ_THAN, r">="],
    [LESS_THAN, r"<"],
    [GREATER_THAN, r">"],
    [NOT_EQUALS, r"!="],
    [EQUALS, r"="],
    [AND, r"y"],
    [OR, r"o"],
    [IDENTIFIER, r"([a-zA-Z]+\d*)+"],
    [FLOAT, r"\d+\.\d+"],
    [INTEGER, r"\d+"],
    [OPENPAR, r"\("],
    [CLOSEPAR, r"\)"]
]

OPERATORS = [LESS_EQ_THAN, LESS_THAN, GREATER_EQ_THAN, GREATER_THAN, EQUALS, NOT_EQUALS]
LOGICAL_OPERATORS = [AND, OR]



from mihto.lexer.lexer import Token
import mihto.lexer.token_types as ttypes
from mihto.parser.nodes import ExpressionNode, VarRefNode, FloatNode, IntegerNode


class Parser:

    def __init__(self, tokens):
        self.tokens = tokens

    def parse(self):
        expression = self._parse_expression()
        return expression

    def _parse_expression(self):
        if self.peek(ttypes.OPENPAR):
            self.consume(ttypes.OPENPAR)
            expression = self._parse_expression()
            self.consume(ttypes.CLOSEPAR)
        else:
            expression = self._parse_atomic_expression()
        return expression

    def _parse_atomic_expression(self) -> ExpressionNode:
        value1 = self._parse_value()
        operator = self._parse_operator()
        value2 = self._parse_value()
        expression = ExpressionNode(value1, operator, value2)
        if self.peek(ttypes.OPERATORS):
            operator2 = self._parse_operator()
            value3 = self._parse_value()
            expression = ExpressionNode(expression, ttypes.AND, ExpressionNode(expression.value2, operator2, value3))

        if self.peek(ttypes.LOGICAL_OPERATORS):
            logical_operator = self._parse_logical_operator()
            expression = ExpressionNode(expression, logical_operator, self._parse_expression())

        return expression

    def _parse_operator(self) -> str:
        if self.peek(ttypes.LESS_EQ_THAN):
            operator = self.consume(ttypes.LESS_EQ_THAN)
        elif self.peek(ttypes.LESS_THAN):
            operator = self.consume(ttypes.LESS_THAN)
        elif self.peek(ttypes.GREATER_EQ_THAN):
            operator = self.consume(ttypes.GREATER_EQ_THAN)
        elif self.peek(ttypes.GREATER_THAN):
            operator = self.consume(ttypes.GREATER_THAN)
        elif self.peek(ttypes.EQUALS):
            operator = self.consume(ttypes.EQUALS)
        else:
            operator = self.consume(ttypes.NOT_EQUALS)
        return operator.type

    def _parse_logical_operator(self) -> str:
        if self.peek(ttypes.AND):
            operator = self.consume(ttypes.AND)
        else:
            operator = self.consume(ttypes.OR)
        return operator.type

    def _parse_value(self):
        if self.peek(ttypes.FLOAT):
            node = self._parse_float()
        elif self.peek(ttypes.INTEGER):
            node = self._parse_integer()
        else:
            node = self._parse_varref()
        return node

    def _parse_varref(self) -> VarRefNode:
        return VarRefNode(self.consume(ttypes.IDENTIFIER).value)

    def _parse_float(self) -> FloatNode:
        return FloatNode(float(self.consume(ttypes.FLOAT).value))

    def _parse_integer(self) -> IntegerNode:
        return IntegerNode(int(self.consume(ttypes.INTEGER).value))

    def consume(self, expected_type: str) -> Token:
        token = self.tokens.pop(0)

        if token.type == expected_type:
            return token
        else:
            raise RuntimeError("Expected token type {} but got {} instead".format(expected_type, token.type))

    def peek(self, expected_type, offset=0) -> bool:
        if type(expected_type) is str:
            return self.tokens and self.tokens[offset].type == expected_type
        else:
            return self.tokens and self.tokens[offset].type in expected_type



"""A setuptools based setup module.
See:
https://packaging.python.org/en/latest/distributing.html
https://github.com/pypa/sampleproject
"""
from setuptools import setup, find_packages

setup(
    name='mihto',
    version='0.0.3',
    description='Easy and simple evaluation language',
    url='https://github.com/OrbitalAds/Mihto',
    author='OrbitalAds',
    author_email='dev@orbitalads.io',
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Developers',
        'Topic :: Software Development :: Build Tools',
        'License :: OSI Approved :: MIT License',
        'Natural Language :: Spanish',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
    ],
    keywords='evaluation evaluate language',
    packages=find_packages(exclude=['docs', 'examples', 'tests'])
)
